REBOL [
	Title: "REBOL/Services - Server Request Processing"
	Author: "Carl Sassenrath"
	Notice: "Confidential source code. Under development."
	Copyright: "REBOL Technology 2005-2008, All rights reserved."
]

;-- Request Handler ----------------------------------------------------------
;
; This section of code is called after the entire request block has been
; successfully loaded and all payload data has been received.
;
; The request format is a sequence of blocks. Each block holds a command
; and its arguments:
;
;   [request seq # ...]
;   [command1 arg1 arg2]
;   [command2 arg1 arg2]
;
; The response format is very similar and can be parsed by the same code:
;
;   done [reply seq # ...]
;   ok   [command1 result1 result2]
;   fail [command1 result1 ...
;
; Each command is done in sequence. Commands can return:
;
;   OK    - success (result is a value or block)
;   FAIL  - a controlled failure (uses a THROW in the command)
;   ERROR - an uncontrolled failure (error throw)
;
; Errors of type 'user are also allowed, and the first argument (string)
; will be returned in the result for the command. If the command is not
; recognized by a service, it must THROW 'bad-command to allow other services
; to have a chance at executing it when multiple services are available.
; For example, doing an INFO in the Home service.

do-request: func [
	"Evaluates all the commands in a request and returns a reply block."
	session [object!]
	request [block!]
	/local req begin cmd error result
][
	session/this-service: none
	begin: now/precise
	either all [
		; Check for request header (includes data like seq number)
		req: request/1
		block? req
		req/1 = 'request
		session/results: reduce ['doing req]
	][
		; Process each request command; build results block:
		foreach command next request [
			either block? command [
				cmd: command/1
				either error? error: try [ ; error set to NONE on success
					catch [result: do-command session command none]
				][
					error: disarm :error
					if error/code = 800 [error/id: error/arg1 error/near: none]
					debug error ['request error/code error/id error/near]
					dump-error error
					append-reply session 'error cmd reduce [error/code error/id error/near]
					break
				][
					either error [
						debug error ['request-fail cmd]
						append-reply session 'fail cmd error
					][
						append-reply session 'ok cmd result
					]
					if error [break] ;if all [error not session/fail-ok] [break]
				]
			][
				debug error ['request-invalid command]
				append-reply session 'fail command 'invalid
			]
		]
		summarize-reply session difference now/precise begin
	][
		; Bad request header. Custom build a reply:
		session/results: compose/deep [
			errors [
				reply service (session/config/service-info/title)
				commands 0 why bad-request-header
			]
		]
	]
	debug reply [session/results]
	session/results
]

do-command: func [
	"Evaluates a command and returns result or throws a failure or error."
	session [object!]
	command [block!]
	/local servs spec result fail name home
][
	debug request ['do-command command]
	servs: reduce ['home find-service session/config 'home]
	name: session/this-service
;	servs: session/config/service-list
	; If the explicit service reference or service focus:
;	if any [
		if path? command/1 [
			name: command/1/1
			change command command/1/2
		]
;		name: session/this-service
;	][
	if name [
		spec: find-service session/config name ; returns service object
		if not spec [bad-service name]
		; Make short list of services. Always include home services:
		insert servs reduce [name spec]
	]
	; Try all services in order (current service first):
	foreach [name spec] servs [
		; Ignore 'bad-command command failures until all services have tried.
		if spec [
			fail: catch [result: spec/dispatch session command none]
			if not fail [return result]
			if fail <> 'bad-command [throw fail] ; relay the throw
		]
	]
	throw 'bad-command
]

summarize-reply: func [
	"Generate summary results and insert them."
	session [object!]
	time
	/local cmds fails out stop
][
	; Tally results:
	cmds: 0
	fails: 0
	foreach [stat result] skip session/results 2 [
		if stat = 'ok [
			cmds: cmds + 1
			if result/1 = 'stop [stop: true]
		]
		if find [error fail] stat [fails: fails + 1  cmds: cmds + 1]
		; (don't count alerts, etc.)
	]

	; Generate the reply header:
	out: session/results/2
	out/1: 'reply
	repend out ['service session/config/service-info/title 'commands cmds 'time time]
	if not zero? fails [repend out ['fail fails]]
	if stop [append out 'stop]
	session/results/1: pick [done errors] zero? fails
	new-line/skip session/results on 2
]

append-reply: func [
	"Append result code and result block onto the reply block."
	session
	code [word!]
	command
	result
][
	repend session/results [code append reduce [command] result]
]

restart?: func [
	"Has client asked server to restart?"
	request [block!]
][
	foreach [stat result] request [
		if all [
			stat = 'ok
			result/1 = 'stop
			find result 'restart
		][
			return true
		]
	]
	false
]
