REBOL [
	Title: "REBOL/Services - CGI Server Handler"
	Author: "Carl Sassenrath"
	Notice: "Confidential source code. Under development."
	Copyright: "REBOL Technology 2005-2008, All rights reserved."
]

;-- CGI Server Handler -------------------------------------------------------
;
;   Direct CGI requests are not handled in the same way as async
;   TCP requests. CGI is processed synchronously. This function
;   returns only when the request has been processed.
;
;   Note that the main code must be run with the -c (cgi) option
;   in order to initialize the system/options/cgi object.
;
;   Also, CGI must disable DEBUG to prevent debug messages
;   from being merged with the result packet. However, the log
;   output of DEBUG is not a problem.

add-server-handler [

	scheme: 'cgi

	opener: func [
		spec
		config [object!]
	][
		debug connect ['opening-cgi-server]
		handler config ; (handle it right now)
		return none
	]

	handler: func [
		config [object!]
		/local cgi data buffer port sid session
	][
		cgi: system/options/cgi
		debug tcp ['handle-cgi-server cgi/request-method cgi/remote-addr cgi/remote-host]

		if cgi/request-method <> "post" [show-service-banner]

		print "content-type: text/plain^/" ; minimal HTTP header

		data: make binary! 1020
		buffer: make binary! 16380
		port: system/ports/input
		sid: none ; session id

		if deny-ip? config cgi/remote-addr [
			debug connect ['deny cgi/remote-addr]
			print {FATAL "access denied"} ; do not use send-packet - no session
			quit
		]

		; Start reading from input port. Find a valid packet header:
		while [positive? read-io port buffer 16378][
			append data buffer
			clear buffer
			if not sid [
				if sid: header-found? data [
					either session: load-session-file config sid [
						if session/remote-ip <> cgi/remote-addr [
							port: system/ports/output
							set-modes port [binary: true] ; send-packet assumes it's a binary port!
							port/user-data: session
							send-fatal port "IP address mismatch"
						]
					][
						session: make-session none config
						session/remote-ip: cgi/remote-addr
						save-session-file config session
					]
					session/recv-buf: data ; use CGI's buffer
					port/user-data: session
				]
			]
		]

		if not session [show-service-text]

		debug tcp ['cgi-read session/id length? session/recv-buf]

		; Process packet (there may be more than one, concatenated)
		; (This code is roughly same as found in server-tcp.r).
		port: system/ports/output
		set-modes port [binary: true] ; send-packet assumes it's a binary port!
		port/user-data: session
		if error? arg: try [
			while [
				all [
					not empty? session/recv-buf
					arg: load-packet session
				]
			][
				if word? arg [send-fatal port arg]
				if block? arg [
					if not empty? session/changed [change-session session]
					arg: do-request session arg
					send-packet port arg
					if not empty? session/changed [change-session session]
				]
			]
		][
			send-fatal port reduce ['error disarm :arg]
		]

		save-session-file config session
		debug tcp ['cgi-done session/id cgi/remote-addr]
	]

	send-fatal: func [
		"Low level packet failure response. (server only)"
		port [port!]
		error-info
	][
		debug error ['fatal error-info]
		send-packet port reduce ['fatal error-info]
		quit
	]

	show-service-banner: does [
		print "content-type: text/html^/" ; minimal HTTP header
		debug tcp ['cgi-get "Browser query - show banner"]
		print service-banner
		if not system/options/cgi/request-method [wait 2]
		quit
	]

	show-service-text: does [
		debug tcp ['cgi-bad "Invalid access"]
		print "REBOL/Services Error:"
		print "Invalid service access - bad packet or session"
		quit
	]

	service-banner: trim/auto {
		<html><head>
		<title>Invalid Access</title>
		<style type="text/css">
		P, BODY, TD {Font-Family : Arial, Sans-serif; Font-size: 10pt}
		</style></head>
		<body>
		<br><br>
		<center>
		<a href="http://www.rebol.com"><img
		src="http://www.rebol.net/graphics/reb-logo.gif" width="176" height="44" border="0"></a>
		<br><b><font size="4">REBOL/Services Site</font></b></p>
		<p>You must use a REBOL/Services client to access this service.</p>
		<p>Go to <a href="http://www.rebol.com">www.REBOL.com</a>
		for more information.</p>
		</center></body></html>
	}

]
