REBOL [
	Title: "REBOL/Services - Async-modes TCP Server Handler (for View 1.3)"
	Author: "Gabriele Santilli"
	Notice: "Confidential source code. Under development."
	Copyright: "REBOL Technology 2005-2008, All rights reserved."
]

;-- TCP Server Handler for non-/async Core -----------------------------------

; uses code from async-protocol.r by Gabriele Santilli, Romano Paolo Tenca and
; Maarten Koopmans

add-server-handler [

	scheme: 'tcp

	wait-find: func [port [port!]][find system/ports/wait-list port]
	wait-stop: func [port [port!]][remove wait-find port]
	wait-start: func [port [port!]][insert tail system/ports/wait-list port]

	opener: func [
		port [port!]
		config [object!]
	][
		debug connect ['opening-tcp-server port/port-id]
		port/user-data: config  ; server itself does not use a session here
		port/async-modes: 'accept
		port/awake: :on-connect
		wait-start port
		open/direct/binary/no-wait port
	]

	on-connect: func [
		listen [port!]
		/local conn session
	] [
		if conn: attempt [first listen] [
			conn/user-data: session: make-session conn listen/user-data ;(config)
			session/remote-ip: conn/remote-ip
			debug connect ['open session/id conn/remote-ip conn/remote-port conn/local-port]
			conn/locals: make object! [
				busy?:          false   ; is true when we are inside on-data, used to avoid recursion
				max-retry:      25      ; max number of connection attempts (none! = forever) -Romano
				retry:          0       ; counter for retry -Romano
				max-transfer:   16 * 1024  ; max TCP transfer size accepted
				;dnsport:        none    ; async dns port
				wakeup?:        none    ; if true, returns from wait
				outBuffer:      make binary! 1024
				close?:         none    ; close port after all data has been sent
			]
			set-modes conn [async-modes: 'read]
			conn/awake: :on-data
			;conn/status: 'open
			wait-start conn
			if deny-ip? session/config conn/remote-ip [
				debug connect ['deny conn/remote-ip]
				send-fatal conn "Access denied"
			]
		]
		false
	]
	
	no-block:       -3      ; return code from IO indicating not to block
	would-block?:   func [err] [517 = get in err 'code] ; checks blocking error

	on-data: func [port /local data err res max-transfer outBuffer] [
		;debug async/locals/busy? "Recursive call to on-data!!! (Or, most likely, it was interrupted last time.)"
		max-transfer: port/locals/max-transfer
		res: false
		port/locals/busy?: true
		; try to receive
		data: make string! max-transfer + 2
		either error? err: try [read-io port data max-transfer] [
			if not would-block? disarm :err [
				attempt [send-fatal port reduce ['tcp-error disarm :err]]
				close-server port
				port/locals/busy?: false
				return false
			]
		] [
			if all [err < -1 err <> no-block] [
				attempt [send-fatal port reduce ['tcp-error reduce ['read-io err]]]
				close-server port
				port/locals/busy?: false
				return false
			]
			if find [0 -1] err [
				; port closed
				debug connect ['close port/user-data/id port/remote-ip port/remote-port port/local-port]
				close-server port
				port/locals/busy?: false
				return false
			]
			if err > 0 [
				append port/user-data/recv-buf data
				debug tcp-read [err 'bytes length? port/user-data/recv-buf 'total]
				if error? err: try [
					while [ ; multiple requests may be concatenated in the buffer
						all [
							not empty? port/user-data/recv-buf
							data: load-packet port/user-data
						]
					][
						if word? data [
							send-fatal port data
						]
						if block? data [
							if not empty? port/user-data/changed [change-session port/user-data]
							data: do-request port/user-data data
							send-packet port data
							if not empty? port/user-data/changed [change-session port/user-data]
						]
					]
				] [
					send-fatal port reduce ['error disarm :err]
				]
				res: found? port/locals/wakeup?
			]
		]
		; try to send
		outBuffer: port/locals/outBuffer
		if not empty? outBuffer [
			either error? err: try [
				write-io port outBuffer min max-transfer length? outBuffer
			] [
				if not would-block? disarm :err [
					print ["*** Server Error: " mold disarm err] 
					close-server port
					port/locals/busy?: false
					return false
				]
			] [
				if all [err < -1 err <> no-block] [
					print ["*** Server Error: write-io aborted (" err ")"] 
					close-server port
					port/locals/busy?: false
					return false
				]
				if err > 0 [
					port/locals/outBuffer: outBuffer: skip outBuffer err
					if tail? outBuffer [
						port/locals/outBuffer: clear head outBuffer
						res: res or found? port/locals/wakeup?
						port/async-modes: 'read
						if port/locals/close? [close-server port]
					]
				]
			]
		]
		port/locals/busy?: false
		res
	]

	close-server: func [
		"Close server side of connection, free the session."
		port
	][
		attempt [close port]
		wait-stop port
		free-session port/user-data
	]

	send-fatal: func [
		"Low level packet failure response. (server only)"
		port [port!]
		error-info
	][
		;print "SEND-FATAL"
		;print mold :error-info
		debug error ['fatal error-info]
		send-packet port reduce ['fatal error-info]
		;attempt [close-server port]
		port/locals/close?: yes
	]

	send-packet: func [
		port [port!]
		request
	][
		insert tail port/locals/outBuffer request: mold-packet port/user-data request
		debug tcp ['send-packet length? request]
		set-modes port [async-modes: [read write]]
	]

	set-wakeup: func [
		port [port!]
		flag [logic!]
	] [
		port/locals/wakeup?: any [port/locals/wakeup? flag]
	]

	clear-wakeup: func [
		port [port!]
	] [
		port/locals/wakeup?: none
	]
]
